--------------------------------------------------------------
--  Subject : Interpreter writing
--  Copyright (c) 1999 by A.Baranovsky
--
--     Language :
--     Program    = "program"  { Statement } "end"
--     Statement  = ( Assignment | Print ) ";"
--     Print      = "print" Expression { "," Expression }
--     Assignment = ID ":=" Expression
--     Expression = Term   { ( "*" | "/" ) Term   }
--     Term       = Factor { ( "+" | "-" ) Factor }
--     Factor     = ID | "(" Expression ")" | - Factor | Number
--     Number     = Digit { Digit }
--     ID         = Letter { ( Letter | Digit ) }
--     Digit      = "0"|"1"|..."9"
--     Letter     = "a"|"b"|..."z"
--------------------------------------------------------------

constant
  WD_PROGRAM = 1
  WD_PRINT   = 2
  WD_END     = 3

  OP_PLUS    = -1
  OP_MINUS   = -2
  OP_DIV     = -3
  OP_MULT    = -4
  OP_ASSIGN  = -5
  OP_MINUS1  = -6
  OP_PRINT   = -7
  OP_HALT    = -8

  OPERS      = [ '+', '-', '*', '/', ':=', 'U-', 'PRINT', 'HALT' ]   

  C_OPER     = 'O'
  C_SPEC     = 'S'
  C_CONST    = 'C'
  C_VAR      = 'V'
  C_KEYWORD  = 'K'

  SP_L       = -101
  SP_R       = -102
  SP_S       = -103
  SP_C       = -104 

  ID         = 1
  CLS        = 2

  Err_program       = 'program expected'
  Err_end           = 'end expected'
  Err_SP_R          = ')  expected'
  Err_SP_S          = ';  expected'
  Err_OP_ASSIGN     = ':= expected'
  Err_var           = 'Variable expected' 
  Err_DivByZero     = 'Division by zero'  
  Err_VarNotDefined = 'Variable is not defined'
end

class TFailrue : Exception
  function Init( InitMessage ) constructor
  function ShowError() 
end

implementation TFailrue

  local Message

  function Init( InitMessage )
    Exception.Create()
    Message := InitMessage 
  return 0

  function ShowError()
    writeln( Message )
  return 0 
end


class TStack 
  function Init() constructor
  function Push( I )
  function Pop( var I )
  function Empty()
end

class TSymbolTable 
  function Init() constructor
  function App( S )
  function Print()
  function GetName( I ) 
  function PutInteger( I, K )
  function GetInteger( I )
  function AppTempVar( K )
  function SetTail()
  function EraseTail()
end

class TScanner  
  function Init( InitBuff, InitSymbolTable ) constructor
  function GetToken() 
end

class TParser 
  function Init( InitScanner, InitPostfix )
  function Parse_PROGRAM() 
  function Parse_PRINT()
  function Parse_ASSIGNMENT() 
  function Parse_EXPRESSION() 
  function Parse_TERM()
  function Parse_FACTOR()
  function Call_SCANNER()
  function App_POSTFIX( I )
  function Test( L, M )
end

class TPostfix  
  function Init( InitSymbolTable ) constructor
  function App( I )
  function Print() 
  function Execute()
end

implementation TStack
  local A, Card
  function Init()
    A := array( 100 ) 
    Card := 0
  return 0
  function Push( I )
    Inc( Card )
    A[ Card ] := I
  return 0 
  function Pop( var I )
    I := A[ Card ]
    Dec( Card )
  return 0
  function Empty()
  return Card = 0 
end


implementation TSymbolTable

  local A, Card, Bound

  function Init()
    A    := array( 100 )
    Card := 0
    App( 'program' )
    App( 'print'   )
    App( 'end'     )
  return 0

  function App( S )
    try
      Found := false
      for I:=1 to Card do
        if S = A[ I, 1 ] then
          raise 0
        end
      end
    except
      Found := true
    end
    if not Found then
      Inc( Card )
      I := Card
      A[ I ] := [ S, * ]
    end
  return I

  function Print()
    writeln( 'Symbol table' )
    for I:=1 to Card do
      if Null( A[ I, 1 ] ) then 
        S1 := 'NIL'
      else
        S1 := A[ I, 1 ]
      end 
      if Null( A[ I, 2 ] ) then
        S2 := 'NIL'
      else
        S2 := STR( A[ I, 2 ] ) 
      end 
      writeln( STR( I ) + '.   ' + S1 + ' ' + S2 )
    end
  return 0

  function GetName( I )
  return A[ I, 1 ]

  function PutInteger( I, K ) 
    A[ I, 2 ] := K
  return 0

  function GetInteger( I )
    if Null( A[ I, 2 ] ) then
      raise TFailrue( Err_VarNotDefined )                            
    end
  return A[ I, 2 ]

  function AppTempVar( K ) 
    Inc( Card )
    A[ Card ] := [ *, K ]                        
  return Card

  function SetTail() 
    Bound := Card
  return 0

  function EraseTail()
    while Card > Bound do
      Destroy( A[ Card ] )
      Dec( Card ) 
    end
  return 0

end


implementation TScanner
  local Buff, P, SymbolTable

  function Init( InitBuff, InitSymbolTable )
    Buff := InitBuff
    P := 0
    object &SymbolTable : TSymbolTable
    SymbolTable == InitSymbolTable
  return 0

  function IsDigit( Ch )
  return ( Ch >= '0' ) and ( Ch <= '9' )

  function IsLetter( Ch )
  return ( Ch >= 'a' ) and ( Ch <= 'z' )

  function GetNextChar()
    Inc( P )
    Ch := Substr( Buff, P, 1 )
  return Ch

  function BackUp()
    Dec( P )
  return 0

  function GetToken()
    try
      repeat
        Ch := GetNextChar()
        if Ch = ' ' then
          --------  continue  -----------
        elsif Ch = '+' then
          Token := [ OP_PLUS, C_OPER ]
          raise Break
        elsif Ch = '-' then 
          Token := [ OP_MINUS, C_OPER ]
          raise Break
        elsif Ch = '*' then  
          Token := [ OP_MULT, C_OPER ]
          raise Break
        elsif Ch = '/' then   
          Token := [ OP_DIV,  C_OPER ]
          raise Break
        elsif Ch = ':' then
          Ch := GetNextChar()
          if Ch = '=' then   
            Token := [ OP_ASSIGN,C_OPER ]
            raise Break
          end
        elsif Ch = '(' then
          Token := [ SP_L,     C_SPEC ]
          raise Break
        elsif Ch = ')' then
          Token := [ SP_R,     C_SPEC ]
          raise Break
        elsif Ch = ';' then 
          Token := [ SP_S,     C_SPEC ]
          raise Break
        elsif Ch = ',' then  
          Token := [ SP_C,     C_SPEC ]
          raise Break
        elsif IsDigit( Ch ) then
          object Stack() : TStack
          S := ''
          while IsDigit( Ch ) do
            S  := S + Ch
            Stack.Push( Ch )
            Ch := GetNextChar()
          end
          BackUp()
          I := SymbolTable.App( S )
          Token := [ I, C_CONST ]
          K := 0
          M := 1 
          while not Stack.Empty() do
            Stack.Pop( Ch )
            D := Ord( Ch ) - Ord( '0' )
            K := K + D * M
            M := M * 10
          end
          SymbolTable.PutInteger( I, K )
          raise Break
        elsif IsLetter( Ch ) then
          S := ''
          while IsLetter( Ch ) or IsDigit( Ch ) do
            S  := S + Ch
            Ch := GetNextChar()
          end
          BackUp()
          I := SymbolTable.App( S )
          if I <= WD_END then
            Token := [ I, C_KEYWORD ]
          else
            Token := [ I, C_VAR ]
          end
          raise Break
        else
          raise TFailrue( 'Syntax error' ) -- implicit constructor call
        end
      until FALSE
    except
      on TFailrue do
        raise
      else
      ----- continue -----------
      end
    end
  return Token
end

implementation TParser

  local CurrToken, Scanner, Postfix

  function Init( InitScanner, InitPostfix )
    object &Scanner : TScanner
    object &Postfix : TPostfix
    Scanner == InitScanner
    Postfix == InitPostfix
  return 0

  function Parse_PROGRAM()
    Call_SCANNER()
    Test( WD_PROGRAM, Err_program ) 
    Call_SCANNER()
    while CurrToken[ ID ] <> WD_END do
      if    CurrToken[ ID ] = WD_PRINT then
        Parse_PRINT()
      elsif CurrToken[ CLS ] = C_VAR   then
        Parse_ASSIGNMENT()
      else
        Test( WD_END, Err_end )
      end
      Test( SP_S, Err_SP_S ) 
      Call_SCANNER()
    end
    App_POSTFIX( OP_HALT )
  return 0

  function Parse_PRINT()
    repeat
      Call_SCANNER()
      Parse_EXPRESSION()
      App_POSTFIX( OP_PRINT )
    until CurrToken[ ID ] <> SP_C
  return 0

  function Parse_ASSIGNMENT()
     App_POSTFIX( CurrToken[ ID ] )
     Call_SCANNER()
     Test( OP_ASSIGN, Err_OP_ASSIGN )   
     Call_SCANNER()
     Parse_EXPRESSION()
     App_POSTFIX( OP_ASSIGN )  
  return 0

  function Parse_EXPRESSION()
    Parse_TERM()
    while ( CurrToken[ ID ] = OP_PLUS  ) or
          ( CurrToken[ ID ] = OP_MINUS ) do
      OP := CurrToken[ ID ] 
      Call_SCANNER()
      Parse_TERM()
      App_POSTFIX( OP )
    end
  return 0

  function Parse_TERM()
    Parse_FACTOR()
    while ( CurrToken[ ID ] = OP_MULT ) or
          ( CurrToken[ ID ] = OP_DIV  ) do
      OP := CurrToken[ ID ]
      Call_SCANNER()
      Parse_FACTOR()
      App_POSTFIX( OP )
    end
  return 0

  function Parse_FACTOR()
    if CurrToken[ ID ] = SP_L then
      Call_SCANNER()
      Parse_EXPRESSION()
      Test( SP_R, Err_SP_R )
      Call_SCANNER()
    elsif CurrToken[ ID ] = OP_MINUS then
      Call_SCANNER()
      Parse_FACTOR()
      App_POSTFIX( OP_MINUS1 )
    elsif ( CurrToken[ CLS ] = C_CONST ) or ( CurrToken[ CLS ] = C_VAR ) then
      App_POSTFIX( CurrToken[ ID ] )
      Call_SCANNER() 
    else
      raise TFailrue( Err_var )
    end
  return 0

  function Call_SCANNER()
    CurrToken := Scanner.GetToken()
  return 0

  function App_POSTFIX( I )
    PostFix.App( I )         
  return 0

  function Test( L, M )
    if CurrToken[ ID ] <> L then
      raise TFailrue( M )        
    end
  return 0
end

implementation TPostfix

  local A, Card, N, Stack, SymbolTable

  function Init( InitSymbolTable )
    object Stack() : TStack 
    object &SymbolTable : TSymbolTable
    SymbolTable == InitSymbolTable
    A := array( 100 )
    Card := 0 
  return 0

  function App( I )
    Inc( Card )
    A[ Card ] := I
  return 0

  function Print()
    writeln( 'Postfix' )

    for I:=1 to Card do
      J := A[ I ]
      if J > 0 then
        S := SymbolTable.GetName( J )
      else
        S := OPERS[ -J ]
      end   
      writeln( STR( I ) + '.   ' + S ) 
    end
  return 0

  function Oper_plus()
    Stack.Pop( I1 )
    Stack.Pop( I2 )
    K1 := SymbolTable.GetInteger( I1 )
    K2 := SymbolTable.GetInteger( I2 )
    I  := SymbolTable.AppTempVar( K1 + K2 )
    Stack.Push( I )
    Inc( N )
  return 0

  function Oper_minus() 
    Stack.Pop( I1 )
    Stack.Pop( I2 )
    K1 := SymbolTable.GetInteger( I1 )
    K2 := SymbolTable.GetInteger( I2 )
    I  := SymbolTable.AppTempVar( K2 - K1 )
    Stack.Push( I )
    Inc( N )
  return 0

  function Oper_mult()  
    Stack.Pop( I1 )
    Stack.Pop( I2 )
    K1 := SymbolTable.GetInteger( I1 )
    K2 := SymbolTable.GetInteger( I2 )
    I  := SymbolTable.AppTempVar( K1 * K2 )
    Stack.Push( I )
    Inc( N )
  return 0

  function Oper_div()   
    Stack.Pop( I1 )
    Stack.Pop( I2 )
    K1 := SymbolTable.GetInteger( I1 )
    K2 := SymbolTable.GetInteger( I2 )

    if K1 = 0 then
      raise TFailrue( Err_DivByZero )
    end

    I  := SymbolTable.AppTempVar( K2 / K1 )
    Stack.Push( I )
    Inc( N )
  return 0

  function Oper_unary()    
    Stack.Pop( I1 )
    K1 := SymbolTable.GetInteger( I1 )
    I  := SymbolTable.AppTempVar( - K1 )
    Stack.Push( I )
    Inc( N )
  return 0

  function Assignment()
    Stack.Pop( I1 )
    Stack.Pop( I2 )
    K1 := SymbolTable.GetInteger( I1 )
    SymbolTable.PutInteger( I2, K1 )
    Inc( N )
  return 0

  function Oper_print()           
    Stack.Pop( I )
    K := SymbolTable.GetInteger( I )
    writeln( K )
    Inc( N )
  return 0

  function Execute() 

    L := array( 8 )
    L[ -OP_PLUS     ] := { || Oper_plus()  }
    L[ -OP_MINUS    ] := { || Oper_minus() }             
    L[ -OP_MULT     ] := { || Oper_mult()  }
    L[ -OP_DIV      ] := { || Oper_div()   }
    L[ -OP_MINUS1   ] := { || Oper_unary() }
    L[ -OP_PRINT    ] := { || Oper_print() }  
    L[ -OP_ASSIGN   ] := { || Assignment() }

    SymbolTable.SetTail()

    N := 1
    repeat
      while A[ N ] > 0 do
        Stack.Push( A[ N ] )
        Inc( N )
      end 

      EvalObject( L[ -A[ N ] ], [] )

      if Stack.Empty() then
        SymbolTable.EraseTail()   
      end

    until A[ N ] = OP_HALT
  return 0

end

object Scanner : TScanner
object SymbolTable : TSymbolTable
object Parser : TParser
object Postfix : TPostfix

-- P := 'program print b; end'              --    variable is not defined
-- P := 'program a := 5 / 0; end'           --    division by zero        
-- P := 'program a 5 / 0; print b ; end'    --    := expected  
P := 'program a := 5 + 6; print a/2,7 ; end'   
writeln( P )

Scanner.Init(  P, SymbolTable )    
SymbolTable.Init()
PostFix.Init( SymbolTable )
Parser.Init(  Scanner, Postfix )

try
  Parser.Parse_PROGRAM()
except
  on E : TFailrue do 
    E.ShowError()  
    raise
  end
end

try
  try
    Postfix.Execute()
  except
    on E : TFailrue do  
      E.ShowError()  
    end
  end 
finally
  Postfix.Print()
  SymbolTable.Print() 
end

return
